# -*- coding: utf-8 -*-
"""
app.api.items
~~~~~~~~~~~~~

This module implements the application API.

:copyright: (c) 2017 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
import sqlalchemy as sa
from flask import (current_app, Blueprint, jsonify, request)
from flask_jwt_extended import create_access_token, jwt_required
from flask_ldap3_login import AuthenticationResponseStatus
from ..extensions import ldap_manager, db
from ..models import Item, Manufacturer, Model, Location, Status
from .. import utils
from ..decorators import jwt_groups_accepted

bp = Blueprint('api', __name__)


def get_generic_model(model):
    # TODO: add pagination
    items = model.query.order_by(model.name)
    data = [item.to_dict() for item in items]
    return jsonify(data)


def create_generic_model(model, mandatory_field='name'):
    data = request.get_json()
    if data is None:
        raise utils.InventoryError('Body should be a JSON object')
    current_app.logger.debug(f'Received: {data}')
    if mandatory_field not in data:
        raise utils.InventoryError(f"Missing mandatory field '{mandatory_field}'", status_code=422)
    try:
        instance = model(**data)
    except TypeError as e:
        message = str(e).replace('__init__() got an ', '')
        raise utils.InventoryError(message)
    db.session.add(instance)
    try:
        db.session.commit()
    except sa.exc.IntegrityError as e:
        db.session.rollback()
        raise utils.InventoryError('IntegrityError', status_code=409)
    return jsonify(instance.to_dict()), 201


@bp.route('/login', methods=['POST'])
def login():
    data = request.get_json()
    if data is None:
        raise utils.InventoryError('Body should be a JSON object')
    try:
        username = data['username']
        password = data['password']
    except KeyError:
        raise utils.InventoryError('Missing mandatory field (username or password)', status_code=422)
    response = ldap_manager.authenticate(username, password)
    if response.status == AuthenticationResponseStatus.success:
        current_app.logger.debug(f'{username} successfully logged in')
        user = ldap_manager._save_user(
            response.user_dn,
            response.user_id,
            response.user_info,
            response.user_groups)
        payload = {'access_token': create_access_token(identity=user.id)}
        return jsonify(payload), 200
    raise utils.InventoryError('Invalid credentials', status_code=401)


@bp.route('/items')
@jwt_required
def get_items():
    # TODO: add pagination
    items = Item.query.order_by(Item._created)
    data = [item.to_dict() for item in items]
    return jsonify(data)


@bp.route('/items', methods=['POST'])
@jwt_required
@jwt_groups_accepted('admin', 'create')
def create_item():
    return create_generic_model(Item, mandatory_field='serial_number')


@bp.route('/items/<item_id>', methods=['PATCH'])
@jwt_required
@jwt_groups_accepted('admin', 'create')
def patch_item(item_id):
    data = request.get_json()
    if data is None:
        raise utils.InventoryError('Body should be a JSON object')
    item = Item.query.get(item_id)
    if item is None:
        raise utils.InventoryError(f'Unknown item id: {item_id}', status_code=422)
    item.name = data.get('name', item.name)
    item.manufacturer = utils.convert_to_model(data.get('manufacturer', item.manufacturer), Manufacturer)
    item.model = utils.convert_to_model(data.get('model', item.model), Model)
    item.location = utils.convert_to_model(data.get('location', item.location), Location)
    item.status = utils.convert_to_model(data.get('status', item.status), Status)
    db.session.commit()
    return jsonify(item.to_dict())


@bp.route('/manufacturers')
@jwt_required
def get_manufacturers():
    return get_generic_model(Manufacturer)


@bp.route('/manufacturers', methods=['POST'])
@jwt_required
@jwt_groups_accepted('admin', 'create')
def create_manufacturer():
    return create_generic_model(Manufacturer)


@bp.route('/models')
@jwt_required
def get_models():
    return get_generic_model(Model)


@bp.route('/models', methods=['POST'])
@jwt_required
@jwt_groups_accepted('admin', 'create')
def create_model():
    return create_generic_model(Model)


@bp.route('/locations')
@jwt_required
def get_locations():
    return get_generic_model(Location)


@bp.route('/locations', methods=['POST'])
@jwt_required
@jwt_groups_accepted('admin', 'create')
def create_locations():
    return create_generic_model(Location)


@bp.route('/status')
@jwt_required
def get_status():
    return get_generic_model(Status)


@bp.route('/status', methods=['POST'])
@jwt_required
@jwt_groups_accepted('admin', 'create')
def create_status():
    return create_generic_model(Status)
