# -*- coding: utf-8 -*-
"""
app.decorators
~~~~~~~~~~~~~~

This module defines some useful decorators.

:copyright: (c) 2017 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
from functools import wraps
from flask_jwt_extended import get_current_user
from .utils import InventoryError


def jwt_groups_required(*groups):
    """Decorator which specifies that a user must have all the specified groups.

    Example::
        @bp.route('/models', methods=['POST'])
        @jwt_required
        @jwt_groups_required('admin', 'create')
        def create_model():
            return create()

    The current user must be in both 'admin' and 'create' groups
    to access this route.

    :param groups: required groups
    """
    def wrapper(fn):
        @wraps(fn)
        def decorated_view(*args, **kwargs):
            user = get_current_user()
            if user is None:
                raise InventoryError('Invalid indentity', status_code=403)
            if not user.is_member_of_all_groups(groups):
                raise InventoryError("User doesn't have the required group(s)", status_code=403)
            return fn(*args, **kwargs)
        return decorated_view
    return wrapper


def jwt_groups_accepted(*groups):
    """Decorator which specifies that a user must have at least one of the specified groups.

    Example::
        @bp.route('/models', methods=['POST'])
        @jwt_required
        @jwt_groups_accepted('admin', 'create')
        def create_model():
            return create()

    The current user must be in either 'admin' or 'create' group
    to access this route.

    :param groups: accepted groups
    """
    def wrapper(fn):
        @wraps(fn)
        def decorated_view(*args, **kwargs):
            user = get_current_user()
            if user is None:
                raise InventoryError('Invalid indentity', status_code=403)
            if not user.is_member_of_one_group(groups):
                raise InventoryError("User doesn't have the required group", status_code=403)
            return fn(*args, **kwargs)
        return decorated_view
    return wrapper
